// dcf V0.6 170225 qrt@qland.de
//
// V0.1     initial version
// V0.3     automatic sec correction
// V0.4     refined sec correction
// V0.5     returned to simple sec correction, added lost signal sync
// V0.6     ticker and timer use

#include "dcf.h"

MicroBit *Dcf::uBit;

uint8_t Dcf::sec;  
uint8_t Dcf::min;
uint8_t Dcf::hou;
uint8_t Dcf::day;
uint8_t Dcf::wday;
uint8_t Dcf::month;
uint8_t Dcf::year;
 
uint8_t Dcf::hsec;  
uint8_t Dcf::hmin;
uint8_t Dcf::hhou;
uint8_t Dcf::hday;
uint8_t Dcf::hwday;
uint8_t Dcf::hmonth;
uint8_t Dcf::hyear;

uint8_t Dcf::status;               
uint8_t Dcf::bit;               
uint8_t Dcf::parity;               

uint32_t Dcf::dt;

Ticker ticker;
Timer timer;

void Dcf::init(MicroBit *_uBit)
{
    uBit = _uBit;
}

void Dcf::receive()
{
    uBit->io.PIN.setPull(PULL);

    #if DCF_LOGIC==NEGATIVE
    uBit->messageBus.listen(DCF_INPUT_PIN, MICROBIT_PIN_EVT_FALL, secPulse, MESSAGE_BUS_LISTENER_IMMEDIATE);
    uBit->messageBus.listen(DCF_INPUT_PIN, MICROBIT_PIN_EVT_RISE, bitPulse, MESSAGE_BUS_LISTENER_IMMEDIATE);     
    #else
    uBit->messageBus.listen(DCF_INPUT_PIN, MICROBIT_PIN_EVT_RISE, secPulse, MESSAGE_BUS_LISTENER_IMMEDIATE);     
    uBit->messageBus.listen(DCF_INPUT_PIN, MICROBIT_PIN_EVT_FALL, bitPulse, MESSAGE_BUS_LISTENER_IMMEDIATE);
    #endif

    uBit->io.P16.eventOn(MICROBIT_PIN_EVENT_ON_EDGE);

    ticker.attach(&service, 1);                             // attach sec ticker
    timer.start();                                          // start sec timer                               

    while(true)
        fiber_sleep(1000);                                  // sleep 
}

void Dcf::service()
{
    incTime();

    if(status&SYNCDAV && min==58 && sec==50)                // no new sync while data not verified                  
        status &= SYNCRES;                                  // hourly sync, lost signal sync after LOSIGT ms
}

void Dcf::secPulse(MicroBitEvent e)
{
    dt = timer.read_ms();                                   // get last timer
    timer.reset();                                          // restart

    if(!(status & SYNCINI)){                                // ignore first pulses after poweron
        if(++parity == IGNOPU)                              // parity used as init counter
            status |= SYNCINI;      
    }       
    else if(!(status & SYNCSEC)){                           // sec sync
        ticker.detach();                                    // detach ticker        
        ticker.attach(&service, 1);                         // attach
        status |= SYNCSEC;      
    }       
    else if(!(status & SYNC59)){                            // sec 59 sync
        if(dt>1950 && dt<2050){                             // 2000 ms +- 50 ms
            hsec = 0;                                       // start data capture
            parity = 0;
            status |= SYNC59;
        }
    }
    else if((status & (SYNCDAR | SYNCDAT)) == SYNCDAR){     // data ready, but not taken
        ticker.detach();                                    // detach ticker        
        takeData();                                         // take data
        ticker.attach(&service, 1);                         // attach ticker
        status |= SYNCDAT;                                  // data taken

        hsec = 0;                                           // start data verify
        parity = 0;
        status &= ~SYNCDAR;             
    }
    else if((status & (SYNCDAR | SYNCDAV)) == SYNCDAR){     // data ready, but not verified                          
        if(verifyData())                                    // verify data
            status &= SYNCRES;                              // restart sync if error
        else
            status |= SYNCDAV;                              // data verified
    }
}

void Dcf::bitPulse(MicroBitEvent e)
{
    if((status & (SYNC59 | SYNCDAR)) == SYNC59){            // SYNC59 done, data not ready
        if(getData())                                       // get data
            status &= SYNCRES;                              // restart sync if error
    }
}

void Dcf::incTime()
{
    if(++sec < 60) 
        return;

    sec = 0;

    if(++min < 60) 
        return;

    min = 0;

    if(++hou < 24) 
        return;

    hou = 0;

    if(++day<29 ||
       (day==29 && month==2 && year%4==0 && year%400!=0) ||
       (day<31 && (month==4 || month==6 || month==9 || month==11)) ||
       (day<32 && (month==1 || month==3 || month==5 || month==7 || month==8 || month==10 || month==12)))
        return;

    day = 1;

    if(++month < 13) 
        return;

    month = 0;
    year++;
}

//         5          4          3          2          1                P parity, Y year, M month, W weekday, D day
// 876543210 9876543210 9876543210 9876543210 9876543210 9676543210     h hour, m minute, L leap second, S time start 
// PYYYYYYYY MMMMMWWWDD DDDDPhhhhh hPmmmmmmmS LZZEARRRRR RRRRRRRRR0     Z zone, A approching zone change, R reserved
// --------  ---------- ---|------ |-------|1                     0 
//
uint8_t Dcf::getData()
{
    dt = timer.read_ms();                                   // bit length

    if(dt>80 && dt<140)                                     // zero
        bit = 0;
    else if(dt>180 && dt<240)                               // one
        bit = 0x80;
    else                                                    // not valid
        return 1;

    if(hsec == 0){                                          // bit 0
        if(bit == 0x80)                                     // always 0
            return 2;             
    }       
    else if(hsec == 20){                                    // bit 20
        if(bit == 0)                                        // always 1
            return 3;             
    }       
    else if(hsec>20 && hsec<29){                            // bits 21-28
        if(hsec < 28){                                      // minute
            hmin = hmin >> 1 | bit;       

            if(hsec==27 && (hmin=decBin(hmin>>1))>59)       
                return 4;    
        }       

        if((parity^=bit) && hsec==28)                       // parity
            return 5;     
    }       
    else if(hsec>28 && hsec<36){                            // bits 29-35
        if(hsec < 35){                                      // hour
            hhou = hhou >> 1 | bit;

            if(hsec==34 && (hhou=decBin(hhou>>2))>24)       
                return 6;     
        }

        if((parity^=bit) && hsec==35)                       // parity 
            return 7;     
    }       
    else if(hsec>35 && hsec<59){                            // bits 36-58
        if(hsec < 42){                                      // day
            hday = hday >> 1 | bit;

            if(hsec==41 && (hday==0 || (hday=decBin(hday>>2))>31))       
                return 8;     
        }
        else if(hsec < 45){                                 // weekday
            hwday = hwday >> 1 | bit;

            if(hsec==44 && (hwday==0 || (hwday>>=5)>7))     
                return 9;     
        }
        else if(hsec < 50){                                 // month
            hmonth = hmonth >> 1 | bit;

            if(hsec==49 && (hmonth==0 || (hmonth=decBin(hmonth>>3))>12))    
                return 10;     
        }
        else if(hsec < 58){                                 // year
            hyear = hyear >> 1 | bit;

            if(hsec==57 && (hyear=decBin(hyear))>99)       
                return 11;    
        }

        if((parity^=bit) && hsec==58)                       // parity
            return 12;

        if(hsec == 58)                                      
            status |= SYNCDAR;                              // data ready
    }

    hsec++;

    return 0;
}

void Dcf::takeData()
{
    sec = 0;  
    min = hmin;
    hou = hhou;
    day = hday;
    wday = hwday;
    month = hmonth;
    year = hyear;        
}

uint8_t Dcf::verifyData()
{    
    return (hou!=hhou || day!=hday || wday!=hwday || month!=hmonth || year!=hyear);
}

uint8_t Dcf::decBin(uint8_t b)
{
    return (b >> 4) * 10 + (b & 0x0f);
}

// uint8_t Dcf::parity(uint8_t b)
// {
//     b ^= b >> 4;
//     b ^= b >> 2;
//     b ^= b >> 1;
    
//     return b & 1;
// }

// uint8_t x = 0;
// uBit->io.P0.setDigitalValue(x++ & 1);
